# coding: utf-8

"""
    DBPreferences - unittests
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2015 by the dbpreferences team, see AUTHORS for more details.
    :created: 2015 by JensDiemer.de
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, print_function

import unittest
import sys
import django

from django.core.exceptions import ValidationError
from django import forms

from django_tools.unittest_utils.print_sql import PrintQueries

from dbpreferences.fields import DictModelField, DictData, DictFormField

from test_project.models import UnittestDictModelFieldModel


class TestDictFiels(unittest.TestCase):
    def test(self):
        d=DictModelField().to_python('''{"foo":"bar"}''')
        self.assertEqual(d, {'foo': 'bar'})
        self.assertIsInstance(d, DictData)

        s = DictModelField().get_prep_value(d)
        self.assertEqual(s, "{'foo': 'bar'}")

        s = DictModelField().get_db_prep_value(d, connection=None, prepared=False)
        self.assertEqual(s, "{'foo': 'bar'}")

        s = DictModelField().get_db_prep_save(d, connection=None)
        self.assertEqual(s, "{'foo': 'bar'}")

        print("OK")

    def test_not_null(self):
        self.assertRaises(
            ValidationError,
            DictModelField().to_python, None
        )

    def test_clean(self):
        f = DictModelField().formfield()
        d = f.clean('''{"foo":"bar"}''')
        self.assertEqual(d, {'foo': 'bar'})

        self.assertIsInstance(f, DictFormField)


class UnittestDictFormFieldForm(forms.Form):
    dict_field = DictFormField()


class TestDictFormField(unittest.TestCase):
    """
    form field for preferences dict
    """
    def test_form_fiels(self):
        f = DictFormField()
        d = f.clean('''{"foo":"bar"}''')
        self.assertEqual(d, {'foo': 'bar'})

    def test_not_null(self):
        self.assertRaises(
            ValidationError,
            DictFormField().clean, None
        )

    def test_error(self):
        self.assertRaises(
            ValidationError,
            DictFormField().clean, "error"
        )

    def test_form(self):
        form = UnittestDictFormFieldForm({"dict_field": '''{"foo":"bar"}'''})
        self.assertTrue(form.is_valid())
        d = form.cleaned_data["dict_field"]
        self.assertEqual(d, {'foo': 'bar'})



class TestDictModelField(django.test.TestCase):
    def setUp(self):
        self.instance = UnittestDictModelFieldModel.objects.create(
            dict_field = {'foo': 'bar'}
        )
        # XXX: Does this delete the cache really?
        self.query_set = UnittestDictModelFieldModel.objects.get_queryset()
        self.query_set._result_cache = None

    def test_instance(self):
        print(self.instance.dict_field, type(self.instance.dict_field))
        self.assertEqual(self.instance.dict_field, {'foo': 'bar'})
        self.assertEqual(self.instance.pk, 1)
        self.assertEqual(UnittestDictModelFieldModel.objects.count(), 1)

    # def test(self):
    #     with PrintQueries("Change object"):
    #         self.instance.dict_field["new"]=1
    #         self.instance.save()
    #         self.assertEqual(self.instance.dict_field, {'foo': 'bar', "new":1})
    #
    #     with PrintQueries("get object"):
    #         print("\n****UnittestDictModelFieldModel.objects.all()")
    #         query_set = UnittestDictModelFieldModel.objects.filter(pk=self.instance.pk)
    #         print("\n****UnittestDictModelFieldModel.objects.all() done")
    #         # new_instance = query_set[0]
    #         print("\n***get dict_field")
    #         # d = new_instance.dict_field
    #         d = query_set.values("dict_field")
    #         print("\n***repr dict_field 1")
    #         print(repr(d), type(d))
    #         print("\n***repr dict_field 2")
    #         d = d[0]["dict_field"]
    #         print(repr(d), type(d))
    #         self.assertEqual(d, {'foo': 'bar', "new":1})

    def test_get1(self): # FIXME: Will fail with Python 3, but works with Python 2 ?!?
        with self.assertNumQueries(2):
            # with PrintQueries("test_get():"):
            self.instance.dict_field["test"] = "test_get()"
            self.instance.save()

            instance = UnittestDictModelFieldModel.objects.get(pk=1)
            d = instance.dict_field
            self.assertEqual(d, {'foo': 'bar', 'test': 'test_get()'})

    @unittest.expectedFailure # FIXME!
    def test_values(self):
        """
        FIXME: Will fail, because of:
            https://code.djangoproject.com/ticket/9619
        """
        with self.assertNumQueries(2):
            # with PrintQueries("test_values(): 2"):
            values = UnittestDictModelFieldModel.objects.all().values("dict_field")
            d = values[0]["dict_field"]
            self.assertEqual(d, {'foo': 'bar'})




#!/usr/bin/env python
# coding: utf-8

"""
    run all unittests
    ~~~~~~~~~~~~~~~~~

    :copyleft: 2011-2015 by django-dbpreferences team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import division, absolute_import

import os
import doctest
import sys

import dbpreferences


SKIP_DIRS = (".settings", ".git", "dist", ".egg-info")
SKIP_FILES = ("setup.py", "test.py")


def get_all_doctests(base_path, verbose=False):
    modules = []
    for root, dirs, filelist in os.walk(base_path, followlinks=True):
        for skip_dir in SKIP_DIRS:
            if skip_dir in dirs:
                dirs.remove(skip_dir) # don't visit this directories

        for filename in filelist:
            if not filename.endswith(".py"):
                continue
            if filename in SKIP_FILES:
                continue

            sys.path.insert(0, root)
            try:
                module = __import__(filename[:-3])
            except ImportError as err:
                if verbose:
                    sys.stderr.write(
                        "\tDocTest import %s error %s\n" % (filename, err)
                    )
            except Exception as err:
                if verbose:
                    sys.stderr.write(
                        "\tDocTest %s error %s\n" % (filename, err)
                    )
            else:
                try:
                    suite = doctest.DocTestSuite(module)
                except ValueError: # has no docstrings
                    continue

                test_count = suite.countTestCases()
                if test_count<1:
                    if verbose:
                        sys.stderr.write(
                            "\tNo DocTests in %r\n" % module.__name__
                        )
                    continue

                if verbose:
                    file_info = module.__file__
                else:
                    file_info = module.__name__
                sys.stderr.write(
                    "\t%i DocTests in %r\n" % (test_count,file_info)
                )
                modules.append(module)
            finally:
                del sys.path[0]

    return modules



def load_tests(loader, tests, ignore):
    sys.stderr.write("\ncollect DocTests:\n")
    path = os.path.abspath(os.path.dirname(dbpreferences.__file__))
    modules = get_all_doctests(
        base_path=path,
        # verbose=True
    )
    for module in modules:
        tests.addTests(doctest.DocTestSuite(module))
    return tests



# coding: utf-8

import os
import sys

sys.path.insert(0, os.path.dirname(__file__))

DEBUG = True

SECRET_KEY = "Only for unittests!"

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ":memory:"
    }
}

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.admin',
    'django.contrib.staticfiles',
    'dbpreferences',
    'test_project',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',

    'dbpreferences.middleware.DBPreferencesMiddleware',

    #'django_tools.middlewares.QueryLogMiddleware.QueryLogMiddleware',
)

SITE_ID = 1

STATIC_URL = "/static/"

ROOT_URLCONF = "test_project.urls"

import dbpreferences, django

TEMPLATE_DIRS = (
    os.path.join(os.path.dirname(dbpreferences.__file__), "templates/"),
    os.path.join(os.path.dirname(django.__file__), "contrib/admin/templates"),
)

CACHE_BACKEND = 'dummy://'


# coding: utf-8

"""
    DBPreferences - models
    ~~~~~~~~~~~~~~~~~~~~~~

    Last commit info:
    ~~~~~~~~~~~~~~~~~
    $LastChangedDate: $
    $Rev: $
    $Author: $

    :copyleft: 2009 by the PyLucid team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import sys
import pprint

from django import forms
from django.db import models
from django.contrib.sites.models import Site
from django.utils.translation import ugettext as _
from django.contrib.auth.models import User, Group
from django.core.exceptions import ValidationError
from django.utils.encoding import python_2_unicode_compatible

from dbpreferences.tools import forms_utils, easy_import, data_eval
from dbpreferences.fields import DictModelField, DictField

# The filename in witch the form should be stored:
PREF_FORM_FILENAME = "preference_forms"


def serialize(data):
    return pprint.pformat(data)

def deserialize(stream):
    return data_eval.data_eval(stream)


class PreferencesManager(models.Manager):
    """ Manager class for Preference model """
    def save_form_init(self, form, site, app_label, form_name):
        """ save the initial form values as the preferences into the database """
        form_dict = forms_utils.get_init_dict(form)
        new_entry = Preference(
            site=site,
            app_label=app_label,
            form_name=form_name,
            preferences=form_dict,
        )
        new_entry.save()
        return new_entry, form_dict

    def get_pref(self, form):
        """
        returns the preferences for the given form
        stores the preferences into the database, if not exist.
        """
        import warnings
        warnings.warn(
            "Preference.objects.get_pref() is deprecated!", DeprecationWarning
        )
        assert isinstance(form, forms.Form), ("You must give a form instance and not only the class!")

        current_site = Site.objects.get_current()
        app_label = form.Meta.app_label
        form_name = form.__class__.__name__

        try:
            db_entry = self.get(site=current_site, app_label=app_label, form_name=form_name)
        except Preference.DoesNotExist:
            # Save initial form values into database
            form_dict = self.save_form_init(form, current_site, app_label, form_name)
        else:
            form_dict = db_entry.preferences
        assert isinstance(form_dict, dict)
        return form_dict



@python_2_unicode_compatible
class Preference(models.Model):
    """
    Plugin preferences
    """
    objects = PreferencesManager()

    id = models.AutoField(primary_key=True, help_text="The id of this preference entry, used for lucidTag")

    site = models.ForeignKey(Site, editable=False, verbose_name=_('Site'))
    app_label = models.CharField(max_length=128, editable=False,
        help_text="app lable, must set via form.Meta.app_label")
    form_name = models.CharField(max_length=128, editable=False,
        help_text="preference form class name")

    preferences = DictModelField(null=False, blank=False, #editable=False,
        help_text="serialized preference form data dictionary")

    createtime = models.DateTimeField(auto_now_add=True, help_text="Create time",)
    lastupdatetime = models.DateTimeField(auto_now=True, help_text="Time of the last change.",)
    lastupdateby = models.ForeignKey(User, editable=False, null=True, blank=True,
        related_name="%(class)s_lastupdateby", help_text="User as last edit the current page.",)

    def clean_fields(self, exclude=None):
        """
        validate preferences dict with the form
        """
        message_dict = {}

        if "preferences" not in exclude:
            form = self.get_form_class()
            f = form(self.preferences)
            f.full_clean()
            if f.is_valid():
                self.preferences = f.cleaned_data
                assert isinstance(self.preferences, dict)
            else:
                errors = []
                for k, v in f._errors.items():
                    errors.append("'%s': '%s'" % (k, ", ".join(v)))
                message_dict["preferences"] = errors

        if message_dict:
            raise ValidationError(message_dict)

    def get_form_class(self):
        """ returns the form class for this preferences item """
        from_name = "%s.%s" % (self.app_label, PREF_FORM_FILENAME)
        form = easy_import.import3(from_name, self.form_name)
        return form

    def __str__(self):
        return u"Preferences for %s.%s.%s" % (self.site, self.app_label, self.form_name)

    class Meta:
        unique_together = ("site", "app_label", "form_name")
        permissions = (("can_change_preferences", "Can change preferences"),)
        ordering = ("site", "app_label", "form_name")
        verbose_name = verbose_name_plural = "preferences"


#-----------------------------------------------------------------------------


try:
    # https://github.com/jedie/django-tools#local-sync-cache
    from django_tools.local_sync_cache.local_sync_cache import LocalSyncCache
except ImportError:
    _USER_SETTINGS_CACHE = {}
else:
    _USER_SETTINGS_CACHE = LocalSyncCache(id="DBPreferences_user_settings")


class UserSettingsManager(models.Manager):
    def get_settings(self, user):
        """ Cached access for getting UserSettings instance and settings. """
        if not user.is_authenticated():
            raise UserSettings.DoesNotExist("No settings for anonymous!")

        try:
            (user_settings_instance, user_settings) = _USER_SETTINGS_CACHE[user.pk]
        except KeyError:
            user_settings_instance = self.get(user=user)
            user_settings = user_settings_instance.settings
            _USER_SETTINGS_CACHE[user.pk] = (user_settings_instance, user_settings)
        assert isinstance(user_settings, dict)
        return user_settings_instance, user_settings

@python_2_unicode_compatible
class UserSettings(models.Model):
    objects = UserSettingsManager()

    user = models.OneToOneField(User, related_name="%(class)s_user")
    settings = DictModelField(null=False, blank=False,
        help_text="serialized user settings data dictionary")

    createtime = models.DateTimeField(auto_now_add=True, help_text="Create time",)
    createby = models.ForeignKey(User, editable=False,
        related_name="%(class)s_createby", help_text="User how has create this entry.",)
    lastupdatetime = models.DateTimeField(auto_now=True, help_text="Time of the last change.",)
    lastupdateby = models.ForeignKey(User, editable=False,
        related_name="%(class)s_lastupdateby", help_text="User how has last edit this entry.",)

    def save(self, *args, **kwargs):
        """ save and update the cache """
        _USER_SETTINGS_CACHE.clear()
        return super(UserSettings, self).save(*args, **kwargs)

    def __str__(self):
        return u"UserSettings for %r: %r" % (self.user, self.settings)

    class Meta:
        ordering = ("user",)
        verbose_name = verbose_name_plural = "User settings"






import unittest
import warnings


class TestOldAPI(unittest.TestCase):
    def test_dict_field(self):
        with warnings.catch_warnings(record=True) as warns:
            warnings.simplefilter('always') # prevent warnings from appearing as errors

            from dbpreferences.fields import DictField
            DictField()

            self.assertEqual(len(warns), 1)
            msg = str(warns[0].message)
            self.assertEqual(msg,
                "You use the old API! DictField was renamed to DictModelField !"
            )


"""
    Data eval
    ~~~~~~~~~

    Evaluate a Python expression string, but only Python data type objects:
        - Constants, Dicts, Lists, Tuples
        - from datetime: datetime and timedelta

    Error class hierarchy:

        DataEvalError
         +-- EvalSyntaxError (compiler SyntaxError)
         +-- UnsafeSourceError (errors from the AST walker)

    A full featured "safe data eval" can be found here:
        https://github.com/newville/asteval/

    :copyleft: 2008-2015 by the dbpreferences team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import ast
import datetime

from django.utils import six

NAME_MAP = {"none": None, "true": True, "false": False}


if six.PY2:
    AST_TEXT_NODES = ast.Str
else:
    AST_TEXT_NODES = (ast.Str, ast.Bytes)


def pprint_node(node):
    for attr in dir(node):
        print(attr, getattr(node, attr, "-"))


class DataEvalError(Exception):
    """ main error class for all data eval errors """
    pass


class EvalSyntaxError(DataEvalError):
    """ compile raised a SyntaxError"""
    pass


class UnsafeSourceError(DataEvalError):
    """ Error class for the SafeEval AST walker """
    pass


class DataEval(object):
    def convert(self, node):
        if isinstance(node, ast.Expression):
            node = node.body

        if isinstance(node, AST_TEXT_NODES):
            return node.s
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Tuple):
            return tuple(map(self.convert, node.elts))
        elif isinstance(node, ast.List):
            return list(map(self.convert, node.elts))
        elif isinstance(node, ast.Set):
            return set(map(self.convert, node.elts))
        elif isinstance(node, ast.Dict):
            return dict((self.convert(k), self.convert(v)) for k, v
                in zip(node.keys, node.values))
        elif not six.PY2 and isinstance(node, ast.NameConstant):
            # ast.NameConstant exists only in Py3 !
            return node.value
        elif isinstance(node, ast.UnaryOp) and \
                isinstance(node.op, (ast.UAdd, ast.USub)) and \
                isinstance(node.operand, (ast.Num, ast.UnaryOp, ast.BinOp)):
            operand = self.convert(node.operand)
            if isinstance(node.op, ast.UAdd):
                return + operand
            else:
                return - operand
        elif isinstance(node, ast.BinOp) and \
                isinstance(node.op, (ast.Add, ast.Sub)) and \
                isinstance(node.right, (ast.Num, ast.UnaryOp, ast.BinOp)) and \
                isinstance(node.left, (ast.Num, ast.UnaryOp, ast.BinOp)):
            left = self.convert(node.left)
            right = self.convert(node.right)
            if isinstance(node.op, ast.Add):
                return left + right
            else:
                return left - right
        elif isinstance(node, ast.Call):
            return self._call(node)
        elif isinstance(node, ast.Name):
            key = node.id
            key = key.lower()
            try:
                return NAME_MAP[key]
            except KeyError as err:
                raise UnsafeSourceError("Name '%s' not supported." % node.id)

        raise UnsafeSourceError(
            "malformed node or string! repr: %r - dump %s" % (
                repr(node), ast.dump(node)
            )
        )

    def _call(self, node):
        try:
            callable_name = node.func.attr
        except AttributeError as err:
            raise UnsafeSourceError(err)

        callable_name = callable_name.lower()

        method_name = "call_%s" % callable_name
        func = getattr(self, method_name, None)
        if func is None:
            raise ValueError("Func '%s' not exists for node: %s" % (
                method_name, repr(node)
            ))
        return func(node)

    def call_datetime(self, node):
        args = list(map(self.convert, node.args))
        return datetime.datetime(*args)

    def call_timedelta(self, node):
        args = list(map(self.convert, node.args))
        return datetime.timedelta(*args)

    def parse(self, source):
        if isinstance(source, dict):
            return source
        elif not isinstance(source, six.string_types):
            raise DataEvalError(
                "source must be string/unicode! (It's type: %r)" % type(source))

        try:
            node = ast.parse(source, mode='eval')
        except SyntaxError as err:
            raise EvalSyntaxError(err)
        except Exception as err:
            raise DataEvalError(err)
        return self.convert(node)


def data_eval(data_string):
    return DataEval().parse(data_string)


if __name__ == '__main__':
    def test(data):
        print("-" * 79)
        data_string = repr(data)
        print(data_string)
        data = DataEval().parse(data_string)
        print(repr(data))
        print("-" * 79)

    print(test("a=1"))
    print(test("foo='bar'"))

    test(datetime.datetime.now())
    test({"dt": datetime.datetime.now()})
    test(datetime.timedelta(seconds=2))



# coding: utf-8


"""
    dbpreferences base form
    ~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2009-2011 by the django-dbpreferences team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import warnings

from django import forms
from django.contrib.sites.models import Site
from django.core.exceptions import ValidationError
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.forms.fields import BooleanField
from django.utils import six

from dbpreferences.models import Preference
from dbpreferences.tools import forms_utils, easy_import


try:
    # https://github.com/jedie/django-tools#local-sync-cache
    from django_tools.local_sync_cache.local_sync_cache import LocalSyncCache
except ImportError:
    _PREFERENCES_CACHE = {}
else:
    _PREFERENCES_CACHE = LocalSyncCache(id="DBPreferences_form")


class DBPreferencesBaseForm(forms.Form):
    preference_cache = {}
    def __init__(self, *args, **kwargs):
        assert(isinstance(self.Meta.app_label, six.string_types))
        super(DBPreferencesBaseForm, self).__init__(*args, **kwargs)

        self.current_site = Site.objects.get_current()
        self.app_label = self.Meta.app_label
        self.form_name = self.__class__.__name__
        self.cache_key = "%s.%s.%s" % (self.current_site.id, self.app_label, self.app_label)

        for name, field in self.fields.items():
            if field.__class__.__name__.startswith("Model"):
                msg = (
                    "Error with field %r from %r:"
                    " Form fields which handle relationships are not supported, yet."
                ) % (name, self.Meta.app_label)
                raise AssertionError(msg)

    def save_form_init(self):
        try:
            Preference.objects.get(
                site=self.current_site, app_label=self.app_label, form_name=self.form_name
            ).delete()
        except Preference.DoesNotExist:
            pass

        # Save initial form values into database
        self.instance, form_dict = Preference.objects.save_form_init(
            form=self,
            site=self.current_site, app_label=self.app_label, form_name=self.form_name
        )

        return form_dict

    def __setitem__(self, key, value):
        if self.data == {}:
            # set self.data with the preferences values
            self.get_preferences()

        self.data[key] = value

    def save(self):
        self.instance.preferences = self.data
        self.instance.save()

    def full_clean(self):
        """
        Fill missing values with initial data and create a warning for that.
        XXX: Is this the best place to do this?
        """
        for name, field in self.fields.items():
            if name not in self.data and field.initial:

                if isinstance(field, BooleanField):
                    # BooleanFields can be missed (unchecked) ;)
                    continue

                # Field doesn't exist in current preferences. Add it if initial value given
                initial = field.initial
                self.data[name] = initial
                msg = (
                    "Use initial value %r for %r cause"
                    " it didn't exist in %s.%s preferences, yet."
                ) % (initial, name, self.app_label, self.form_name)
                warnings.warn(msg)

        super(DBPreferencesBaseForm, self).full_clean()

    def get_preferences(self):
        """
        return current preferences
            1. get the dbpreferenced data from database
            2. validate them
            3. return cleaned_data dict
        """
        try:
            self.instance = self.get_db_instance()
        except Preference.DoesNotExist:
            self.data = self.save_form_init()
        else:
            self.data = self.instance.preferences

        # Cleans all of self.data and populates self._errors and self.cleaned_data
        self.is_bound = True
        self.full_clean()

        if not self.is_valid():
            errors = []
            for item, msg in self._errors.items():
                if item in self.data:
                    value_info = "(current value is: %r)" % self.data[item]
                else:
                    value_info = "(key not set in dict!)"

                errors.append(
                    "'%s' %s: '%s'" % (
                        item, value_info, ", ".join(msg)
                    )
                )

            error_msg = ", ".join(errors)
            raise ValidationError(
                "DBpreferences data for '%s.%s' not valid: %s" % (
                    self.app_label, self.form_name, error_msg
                )
            )

        assert isinstance(self.cleaned_data, dict)
        return self.cleaned_data

    def get_db_instance(self):
        """ returns the database entry instance """
        try:
            self.instance = _PREFERENCES_CACHE[self.cache_key]
        except KeyError:
            self.instance = Preference.objects.get(
                site=self.current_site, app_label=self.app_label, form_name=self.form_name
            )
            _PREFERENCES_CACHE[self.cache_key] = self.instance

        return self.instance


@receiver(post_save, sender=Preference)
def clear_cache(sender, **kwargs):
    _PREFERENCES_CACHE.clear()



# coding: utf-8

"""
    DBPreferences - fields
    ~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2009-2011 by the dbpreferences team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

from __future__ import absolute_import, print_function

import warnings

import sys
import pprint

from django import forms
from django.db import models
from django.contrib.sites.models import Site
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.models import User, Group
from django.utils import six

from dbpreferences.tools import forms_utils, easy_import, data_eval
from dbpreferences.tools.data_eval import DataEvalError



class DictFormField(forms.CharField):
    """
    form field for preferences dict
    
    >>> DictFormField().clean('''{"foo":"bar"}''') == {'foo': 'bar'}
    True
    """
    def clean(self, value):
        """
        validate the form data
        FIXME: How can we get the pref form class for validating???
        """
        value = super(DictFormField, self).clean(value)
        if value == u'':
            # empty value with a required=False
            return u''

        try:
            return DictData(value)
        except DataEvalError as err:
            msg = "Can't deserialize: %s" % err
            raise forms.ValidationError(msg)


class DictData(dict):
    """
    Can init with a dict:
    >>> DictData({"foo":"bar"}) == {'foo': 'bar'}
    True
    
    Can also init with a string:
    >>> DictData('''{"foo":"bar"}''') == {'foo': 'bar'}
    True
    """
    def __init__(self, value):
        if isinstance(value, dict):
            super(DictData, self).__init__(value)
        else:
            d = data_eval.data_eval(value)
            assert isinstance(d, dict)
            super(DictData, self).__init__(d)

        # if isinstance(value, six.string_types):
        #     self.value = value
        #     super(DictData, self).__init__()
        # elif isinstance(value, dict):
        #     self.value = None
        #     super(DictData, self).__init__(value)
        # else:
        #     raise TypeError("init data is not from type str/basestring or dict (It's type: %r)" % type(value))

    def __repr__(self):
        """ used in django admin form field and in DictModelField.get_db_prep_save() """
        return pprint.pformat(dict(self))


# https://docs.djangoproject.com/en/1.8/releases/1.8/#subfieldbase
@six.add_metaclass(models.SubfieldBase) # SubfieldBase will be removed in Django 2.0
class DictModelField(models.Field):
    description = "DictModelField"

    """
    A dict field.
    Stores a python dict into a text field.

    FIXME: Will not work with .values(...), because of:
        https://code.djangoproject.com/ticket/9619

    https://docs.djangoproject.com/en/1.8/howto/custom-model-fields/#converting-values-to-python-objects
    
    >>> d=DictModelField().to_python('''{"foo":"bar"}''')
    >>> d == {'foo': 'bar'}
    True
    >>> isinstance(d, DictData)
    True
    >>> DictModelField().get_db_prep_save(d) == "{'foo': 'bar'}"
    True

    >>> f = DictModelField().formfield()
    >>> f.clean('''{"foo":"bar"}''') == {'foo': 'bar'}
    True
    """

    def get_internal_type(self):
        return "TextField"

    # def __getattribute__(self, item):
    #     result = super(DictModelField, self).__getattribute__(item)
    #     print("DictModelField.getattribute: %20s: %r" % (item, result))
    #     return result
    #
    # def __getattr__(self, item):
    #     result = super(DictModelField, self).__getattr__(item)
    #     print("DictModelField.getattr: %20s: %r" % (item, result))
    #     return result

    def to_python(self, value):
        """
        Converts the input value into the expected Python data type, raising
        django.core.exceptions.ValidationError if the data can't be converted.
        Returns the converted value. Subclasses should override this.

        to_python() will be not called since 1.8:
        https://docs.djangoproject.com/en/1.8/releases/1.8/#subfieldbase
        """
        result = self.from_db_value(value)
        # print("\nDictModelField.to_python() return", result, type(result))
        return result

    def from_db_value(self, value, expression=None, connection=None, context=None):
        """
        Converts the input value into the expected Python data type, raising
        django.core.exceptions.ValidationError if the data can't be converted.
        Returns the converted value. Subclasses should override this.

        from_db_value() is new in new in django 1.8!
        """
        # print("\nDictModelField.from_db_value()")
        if isinstance(value, dict):
            # print("is a dict")
            return value

        if value is None:
            if self.null == False:
                raise forms.ValidationError(_("This field cannot be null."))
            else:
                # print("is None")
                return None

        try:
            result = DictData(value)
            # print("\nDictModelField.from_db_value() return", result, type(result))
            return result
        except DataEvalError as err:
            msg = "Can't deserialize %r: %s" % (value, err)
            raise forms.ValidationError(msg)

    def get_prep_value(self, value):
        """
        Perform preliminary non-db specific value checks and conversions.
        """
        if value is not None:
            return repr(value)
        else:
            if self.null == False:
                raise forms.ValidationError(_("This field cannot be null."))
            else:
                return None

    def formfield(self, **kwargs):
        # Always use own form field and widget:
        kwargs['form_class'] = DictFormField
        return super(DictModelField, self).formfield(**kwargs)


def DictField(*args, **kwargs):
    "TODO: remove this old API support in future!"
    warnings.warn(
        "You use the old API! DictField was renamed to DictModelField !",
        FutureWarning,
        stacklevel=2
    )
    return DictModelField(*args, **kwargs)



"""
    unittests for data eval
    ~~~~~~~~~~~~~~~~~~~~~~~

    :copyleft: 2015 by the dbpreferences team, see AUTHORS for more details.
    :created: 2015 by JensDiemer.de
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import datetime
import unittest


from dbpreferences.tools.data_eval import DataEval, DataEvalError, \
    UnsafeSourceError, EvalSyntaxError


class TestDataEval(unittest.TestCase):
    def literal_eval(self, data_string):
        return DataEval().parse(data_string)

    def assert_eval(self, data):
        data_string = repr(data)
        result = self.literal_eval(data_string)
        self.assertEqual(result, data)

    def test_none(self):
        self.assert_eval(None)

    def test_bool(self):
        self.assert_eval(True)
        self.assert_eval(False)
        self.assert_eval([True, False])
        self.failUnlessEqual(self.literal_eval("true"), True)
        self.failUnlessEqual(self.literal_eval("TRUE"), True)

    def test_const(self):
        self.assert_eval(1)
        self.assert_eval(1.01)
        self.assert_eval("FooBar")
        self.assert_eval(u"FooBar")

    def test_negative_values(self):
        self.assert_eval(-1)
        self.assert_eval(-2.02)

    def test_tuple(self):
        self.assert_eval(())
        self.assert_eval((1, 2))
        self.assert_eval(("1", u"2", None, True, False))

    def test_list(self):
        self.assert_eval([])
        self.assert_eval([1, 2, -3, -4.41])
        self.assert_eval(["foo", u"bar", None, True, False])

    def test_dict(self):
        self.assert_eval({})
        self.assert_eval({1: 2, "a": "b", u"c": "c", "d": -1, "e": -2.02})
        self.assert_eval({"foo": "bar", u"1": None, 1: True, 0: False})

    def test_datetime(self):
        self.assert_eval(datetime.datetime.now())
        self.assert_eval({"dt": datetime.datetime.now()})
        self.assert_eval(datetime.timedelta(seconds=2))

    def test_line_endings(self):
        self.literal_eval("\r\n{\r\n'foo'\r\n:\r\n1\r\n}\r\n")
        self.literal_eval("\r{\r'foo'\r:\r1\r}\r")

    def test_no_string(self):
        self.assertRaises(DataEvalError, self.literal_eval, 1)

    def test_quote_err(self):
        self.assertRaises(UnsafeSourceError, self.literal_eval, "a")
        self.assertRaises(DataEvalError, self.literal_eval, "a")

    def test_unsupported_err(self):
        self.assertRaises(UnsafeSourceError, self.literal_eval, "a+2")
        self.assertRaises(UnsafeSourceError, self.literal_eval, "eval()")
        self.assertRaises(DataEvalError, self.literal_eval, "eval()")

    def test_syntax_error(self):
        self.assertRaises(EvalSyntaxError, self.literal_eval, ":")
        self.assertRaises(EvalSyntaxError, self.literal_eval, "import os")
        self.assertRaises(DataEvalError, self.literal_eval, "import os")


if __name__ == '__main__':
    unittest.main()



# coding: utf-8
"""
    Unittest for DBpreferences

    INFO: dbpreferences should be exist in python path!
"""

from django.utils import six

if __name__ == "__main__":
    # run unittest directly
    import os
    os.environ["DJANGO_SETTINGS_MODULE"] = "test_settings"

from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.db.models import signals

from django_tools.unittest_utils.unittest_base import BaseTestCase

from dbpreferences import models, forms
from dbpreferences.fields import DictModelField, DictData, DictFormField
from dbpreferences.forms import DBPreferencesBaseForm
from dbpreferences.middleware import SettingsDict
from dbpreferences.models import Preference, UserSettings

from test_project.preference_forms import UnittestForm, TestModelChoiceForm



class FormWithoutMeta(DBPreferencesBaseForm):
    pass

class FormMetaWithoutAppLabel(DBPreferencesBaseForm):
    class Meta:
        pass


class GetItemCountDict(dict):
    def __init__(self):
        self.cache_hit = 0
        super(GetItemCountDict, self).__init__()

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)

    def __getitem__(self, key):
        self.cache_hit += 1
        return dict.__getitem__(self, key)


class TestDBPref(BaseTestCase):
    def setUp(self):
        Preference.objects.all().delete()

        forms._PREFERENCES_CACHE = GetItemCountDict()

    def test_form_without_meta(self):
        self.failUnlessRaises(AttributeError, FormWithoutMeta)

    def test_form_meta_without_app_label(self):
        self.failUnlessRaises(AttributeError, FormMetaWithoutAppLabel)

    def test_form_api(self):
        form = UnittestForm()

        # Frist time, the data would be inserted into the database
        self.assertEqual(Preference.objects.count(), 0)
        pref_data = form.get_preferences()
        self.assertEqual(Preference.objects.count(), 1)
        self.failUnless(isinstance(pref_data, dict),
            "It's not dict, it's: %s - %r" % (type(pref_data), pref_data))
        self.failUnlessEqual(pref_data,
            {'count': 10, 'foo_bool': True, 'font_size': 0.7, 'subject': 'foobar'})

        form = UnittestForm()
        self.assertEqual(Preference.objects.count(), 1)
        pref_data = form.get_preferences()
        self.assertIsInstance(pref_data, dict)
        self.assertEqual(Preference.objects.count(), 1)
        self.failUnless(isinstance(pref_data, dict),
            "It's not dict, it's: %s - %r" % (type(pref_data), pref_data))
        self.failUnlessEqual(pref_data,
            {'count': 10, 'foo_bool': True, 'font_size': 0.7, 'subject': 'foobar'})

        # Change a value
        form["count"] = 20
        form.save()

        # Change a value without 
        form = UnittestForm()
        form["foo_bool"] = False
        form.save()

        # Check the changes value
        form = UnittestForm()
        pref_data = form.get_preferences()
        self.failUnlessEqual(pref_data,
            {'count': 20, 'foo_bool': False, 'font_size': 0.7, 'subject': 'foobar'})

    def test_admin_edit(self):
        # Create one db entry
        form = UnittestForm()
        pref_data = form.get_preferences()
        self.assertEqual(Preference.objects.count(), 1)

        pk = Preference.objects.all()[0].pk
        url = reverse("admin:dbpref_edit_form", kwargs={"pk":pk})

        self.create_testusers()
        self.login(usertype="staff")

        response = self.client.get(url)
        self.failUnlessEqual(response.status_code, 200)
        self.assertResponse(response,
            must_contain=("Change Preference", "test_project.UnittestForm"),
            must_not_contain=("Error", "Traceback")
        )

        response = self.client.post(url, data={
            "subject": "new content", "count": 5, "font_size": 1, "_save":"Save"})
        self.failUnlessEqual(response.status_code, 302)

        response = self.client.get(url)
        self.failUnlessEqual(response.status_code, 200)
        self.assertResponse(response,
            must_contain=("Change Preference", "test_project.UnittestForm"),
            must_not_contain=("Error", "Traceback")
        )

        form = UnittestForm()
        self.assertEqual(Preference.objects.count(), 1)
        pref_data = form.get_preferences()
        self.assertEqual(Preference.objects.count(), 1)
        self.failUnless(isinstance(pref_data, dict),
            "It's not dict, it's: %s - %r" % (type(pref_data), pref_data))
        self.failUnlessEqual(pref_data,
            {'count': 5, 'foo_bool': False, 'font_size': 1.0, 'subject': u"new content"})

    def test_model_choice(self):
        """
        Test the form check: Form fields which handle relationships are not supported, yet.
        """
        self.failUnlessRaises(AssertionError, TestModelChoiceForm)

        # A test can look like this, if ModelChoiceField are supported:
#        form = TestModelChoiceForm()
#        pref_data = form.get_preferences()
#
#        pref_obj = Preference.objects.all()[0]
#        form["model_choice"] = pref_obj
#        form.save()
#
#        form = TestModelChoiceForm()
#
#        self.failUnlessEqual(form["model_choice"], pref_obj)


# ----------------------------------------------------------------------------


class TestDictFieldForm(BaseTestCase):
    """ Tests for dbpreferences.fields objects. """
    def test_data_eval1(self):
        d = DictModelField().to_python('''{"foo":"bar"}''')
        self.failUnlessEqual(d, {'foo': 'bar'})
        self.failUnless(isinstance(d, DictData))

    def test_repr1(self):
        """ use get_prep_value() with DictData instance """
        d = DictData({'foo': 'bar'})
        s = DictModelField().get_prep_value(d)
        self.failUnlessEqual(s, "{'foo': 'bar'}")

    def test_repr2(self):
        """ use get_prep_value() with normal dict object """
        s = DictModelField().get_prep_value({'foo': 'bar'})
        self.failUnlessEqual(s, "{'foo': 'bar'}")

    def test_to_python_cant_empty1(self):
        d = DictModelField()
        self.failUnlessRaises(ValidationError, d.to_python, None)

    def test_to_python_cant_empty2(self):
        d = DictModelField(blank=True, null=False)
        self.failUnlessRaises(ValidationError, d.to_python, None)

    def test_to_python_can_empty(self):
        d = DictModelField(blank=True, null=True)
        self.failUnlessEqual(d.to_python(None), None)

    def test_get_prep_value_cant_empty1(self):
        d = DictModelField()
        self.failUnlessRaises(ValidationError, d.get_prep_value, None)

    def test_get_prep_value_cant_empty2(self):
        d = DictModelField(blank=True, null=False)
        self.failUnlessRaises(ValidationError, d.get_prep_value, None)

    def test_get_prep_value_can_empty(self):
        d = DictModelField(blank=True, null=True)
        self.failUnlessEqual(d.get_prep_value(None), None)

    def test_formfield(self):
        d = DictModelField()
        f = d.formfield()
        self.failUnless(isinstance(f, DictFormField))

    def test_formfield_clean(self):
        f = DictFormField()
        d = f.clean('''{"foo":"bar"}''')
        self.failUnlessEqual(d, {'foo': 'bar'})

    def test_formfield_clean_cant_empty(self):
        f = DictFormField()
        self.failUnlessRaises(ValidationError, f.clean, None)

    def test_formfield_clean_can_empty(self):
        f = DictFormField(required=False)
        s = f.clean(None)
        self.failUnlessEqual(s, u'')


# ----------------------------------------------------------------------------


class TestUserSettings(BaseTestCase):
    def setUp(self):
        self.create_testusers()

        self._saved = 0
        self._init = 0
        signals.post_save.connect(self._post_save_handler, sender=UserSettings)
        signals.pre_init.connect(self._post_init_handler, sender=UserSettings)

        models._USER_SETTINGS_CACHE = GetItemCountDict()
        UserSettings.objects.all().delete()

    def _post_save_handler(self, **kwargs):
        self._saved += 1
    def _post_init_handler(self, **kwargs):
        self._init += 1

    def test_low_level(self):
        self.failUnlessEqual(len(models._USER_SETTINGS_CACHE), 0)

        user = self._get_user(usertype="staff")
        user_settings = SettingsDict(user)

        # .get set the value, if not exist 
        self.failUnlessEqual(user_settings.get("Foo", "initial value"), "initial value")
        self.failUnlessEqual(user_settings["Foo"], "initial value")

        # change the existing value
        user_settings["Foo"] = "new value"
        # Now we should get the new value:
        self.failUnlessEqual(user_settings.get("Foo", "initial value"), "new value")

        # before we save, nothing should be created
        self.failUnless(UserSettings.objects.count() == 0)

        self.failUnlessEqual(self._init, 0)
        self.failUnlessEqual(self._saved, 0)
        self.failUnlessEqual(len(models._USER_SETTINGS_CACHE), 0)

        user_settings.save() # increment: pre_init + post_save

        self.failUnlessEqual(self._init, 1)
        self.failUnlessEqual(self._saved, 1)
        self.failUnlessEqual(len(models._USER_SETTINGS_CACHE), 0)
        self.failUnless(UserSettings.objects.count() == 1)

        instance = UserSettings.objects.get(user=user) # increment: pre_init
        self.failUnlessEqual(self._init, 2)
        self.failUnlessEqual(instance.settings, {"Foo": "new value"})

        # Now we should get a cached version, so pre_init should not be increment
        user_settings = SettingsDict(user)
        self.failUnlessEqual(user_settings["Foo"], "new value")

        self.failUnlessEqual(self._init, 3)
        self.failUnlessEqual(self._saved, 1)
        self.failUnlessEqual(models._USER_SETTINGS_CACHE.cache_hit, 2)

    def test_load(self):
        """
        The load() method in middleware.SettingsDict would be called on every get, getitem, etc.
        But the real loading should only done one time.
        """
        user = self._get_user(usertype="staff")
        user_settings = SettingsDict(user)
        self.failUnlessEqual(models._USER_SETTINGS_CACHE.cache_hit, 0)
        user_settings.get("foo", "bar")
        user_settings.get("foo", "bar")
        user_settings["foo"]
        user_settings["foo"]
        self.failUnlessEqual(models._USER_SETTINGS_CACHE.cache_hit, 1)

    def test_view_base(self):
        url = reverse("test_user_settings", kwargs={"test_name": "base_test", "key": "Foo", "value": "Bar"})
        response = self.client.get(url)
        self.failUnlessEqual(response.status_code, 200)
        self.failUnlessEqual(six.text_type(response.content, "UTF-8"), "FooBar")

    def test_view_get(self):
        self.login(usertype="staff")

        # first access via .get() -> returned the default value "initial value"
        url = reverse("test_user_settings",
            kwargs={"test_name": "get", "key": "Foo", "value": "initial value"}
        )
        response = self.client.get(url)
        self.failUnlessEqual(response.status_code, 200)
        self.failUnlessEqual(six.text_type(response.content, "UTF-8"), "initial value")

        # change with .setitem()
        url = reverse("test_user_settings",
            kwargs={"test_name": "setitem", "key": "Foo", "value": "new value"}
        )
        response = self.client.get(url)
        self.failUnlessEqual(response.status_code, 200)
        self.failUnlessEqual(six.text_type(response.content, "UTF-8"), "new value")

        # now, the .get() method should returned the current value
        url = reverse("test_user_settings",
            kwargs={"test_name": "get", "key": "Foo", "value": "initial value"}
        )
        response = self.client.get(url)
        self.failUnlessEqual(response.status_code, 200)
        self.failUnlessEqual(six.text_type(response.content, "UTF-8"), "new value")

    def test_user_settings_cache(self):
        self.login(usertype="staff")
        for no in range(10):
            url = reverse("test_user_settings_cache", kwargs={"no": no})
            response = self.client.get(url)
            self.failUnlessEqual(response.status_code, 200)
            self.failUnlessEqual(six.text_type(response.content, "UTF-8"), str(no))

        #self.failUnlessEqual(self._init, 4)
        self.failUnlessEqual(self._saved, 2)
        self.failUnlessEqual(models._USER_SETTINGS_CACHE.cache_hit, 10)

    def test_anonymous(self):
        """
        The settings would be not saved for anonymous users.
        We allways get the default value back.
        """
        user = AnonymousUser()

        user_settings = SettingsDict(user)
        user_settings["Foo"] = "bar"
        # in a request, we get values set in the past
        self.failUnlessEqual(user_settings["Foo"], "bar")

        # For anonymous user, save() does nothing:
        user_settings.save()
        self.failUnlessEqual(self._saved, 0)

        # In "the next request" we can't get the old value "bar"
        user_settings = SettingsDict(user)
        self.failUnlessEqual(user_settings.get("Foo", "not the initial value"), "not the initial value")

    def test_issues1(self):
        """ https://github.com/jedie/django-dbpreferences/issues/1 """
        user = self._get_user(usertype="staff")
        user_settings = SettingsDict(user)
        user_settings["Foo"] = "Bar"
        user_settings.save()

        models._USER_SETTINGS_CACHE = GetItemCountDict()

        user_settings_instance, user_settings = UserSettings.objects.get_settings(user)
        self.failUnless(isinstance(user_settings_instance, UserSettings))
        self.failUnlessEqual(user_settings, {"Foo":"Bar"})





if __name__ == "__main__":
    # Run this unittest directly
    from django.core import management
    management.call_command('test', 'dbpreferences')
#    management.call_command('test', 'dbpreferences.TestUserSettings.test_issues1')
